"""Utilities for managing concurrency in a sane way."""

from contextlib import contextmanager
import sys
import threading


class Callback(object):

    """
    A threading-based callback supporting both values and exceptions.

    This object mimics the Python coroutine object in supporting :meth:`send`
    and :meth:`throw`, only sending values and throwing exceptions across
    thread boundaries as opposed to between coroutines. It should be used for
    synchronization where more than a typical ``threading.Event`` is needed,
    because values/exceptions need to be associated with those events.

    The :meth:`spawn` and :meth:`die` methods, and the :attr:`event_class`
    attribute can be overridden in subclasses to support alternative,
    thread-like concurrency models (e.g. gevent, eventlet, etc.).
    """

    event_class = threading.Event

    def __init__(self):
        self.event = self.event_class()
        self.value = None
        self.exc_info = None

    def spawn(self, func, args=(), kwargs=None, **opts):

        """
        Abstract method for spawning a function in a new thread.

        The returned object should represent the thread, and have at least a
        ``join()`` method.
        """

        thread = threading.Thread(target=func, args=args,
                                  kwargs=(kwargs or {}))
        thread.daemon = opts.get('daemon', False)
        thread.start()
        return thread

    def reset(self):
        """Reset this callback, allowing it to be used again."""

        self.event.clear()
        self.value = None
        self.exc_info = None

    def set(self):
        """Alias for ``send(None)``."""

        self.send(None)

    def send(self, value):
        """Pass the provided value to the waiting thread."""

        self.value = value
        self.event.set()

    def throw(self, *exc_info):

        """
        Throw the given exception in the waiting thread.

        There are three valid ways of calling this method:

        * No arguments: use ``sys.exc_info()`` as the source of the exception.
        * 1 argument: an exception object. This is raised and immediately
          caught so as to capture a traceback.
        * 3 arguments: interpreted as the three parts of a ``sys.exc_info()``
          tuple.
        """

        if len(exc_info) == 0:
            exc_info = sys.exc_info()
        elif len(exc_info) == 1:
            try:
                raise exc_info[0]
            except Exception:
                exc_info = sys.exc_info()
        elif len(exc_info) == 3:
            pass
        else:
            raise TypeError("Invalid exception argument: %r" % (exc_info,))

        self.exc_info = exc_info
        self.event.set()

    @property
    def catch_exceptions(self):
        """A context manager to catch and throw unhandled exceptions."""

        @contextmanager
        def exception_catcher(die=True):
            try:
                yield
            except Exception, exc:
                self.throw()
                if die:
                    self.die()
                else:
                    raise
        return exception_catcher

    def die(self):
        """Abstract method to kill the current thread."""

        raise SystemExit

    def wait(self):
        """Wait for the callback to be called, and return/raise."""

        self.event.wait()
        if self.exc_info:
            raise self.exc_info[1]
        return self.value


class DummyCallback(Callback):

    """A null callback which does nothing."""

    event_class = object

    def throw(self, *args):
        pass

    def send(self, value):
        pass

    def wait(self):
        pass

    def die(self):
        pass


try:
    import gevent.event
except ImportError:
    pass
else:
    class GeventCallback(Callback):
        """A callback class supporting gevent instead of threading."""

        event_class = gevent.event.Event
        die_exception = gevent.GreenletExit

        def spawn(self, func, args=(), kwargs=None, **opts):
            return gevent.spawn(func, *args, **(kwargs or {}))



from contextlib import closing, nested
from Queue import Queue
import threading
import uuid

import logbook
import zmq

from zrpc.concurrency import Callback, DummyCallback
from zrpc.loadbal import LoadBalancer
from zrpc.server import Server


logger = logbook.Logger('zrpc.multiserver')


class MultiServer(object):

    """A multi-threaded ZMQ server."""

    def __init__(self, addr, registry, connect=False, context=None):
        self.context = context or zmq.Context.instance()
        self.addr = addr
        self.connect = connect
        self.registry = registry

    def run_device(self, callback=DummyCallback(), device=zmq.device):

        """
        Run the ZMQ queue device in a background thread.

        The return value of this function will be the address of the device's
        output socket (an `inproc://`-type address). In order to kill the
        background thread, you need to terminate the context the device is
        attached to.
        """

        output_addr = 'inproc://%s' % uuid.uuid4()
        loadbal = LoadBalancer(self.addr, output_addr, context=self.context)
        loadbal_thread = callback.spawn(loadbal.run,
                                        kwargs={'callback': callback,
                                                'device': device})
        return output_addr

    def run(self, n_workers, callback=DummyCallback(), device=zmq.device):
        loadbal_callback = type(callback)()
        loadbal_addr = self.run_device(callback=loadbal_callback,
                                       device=device)
        # We need the load balancer to be bound before continuing.
        loadbal_callback.wait()

        with callback.catch_exceptions():
            server = Server(loadbal_addr, self.registry, connect=True,
                            context=self.context)
            server_threads = []
            server_callbacks = []

            for i in xrange(n_workers):
                server_callback = type(callback)()
                server_thread = server_callback.spawn(
                    server.run, kwargs={'callback': server_callback})
                server_threads.append(server_thread)
                server_callbacks.append(server_callback)

            # Wait for all callbacks to complete before triggering ours.
            server_sockets = [server_callback.wait()
                            for server_callback in server_callbacks]

        callback.send(server_sockets)

        return loadbal_addr, server_threads



import uuid

from bson import BSON
import zmq


class Error(Exception):

    _class_cache = {}

    class __metaclass__(type):

        def _get_error_cls(cls, error_type):
            generic_type = error_type.split('.')[-1]
            if generic_type not in cls._class_cache:
                cls._class_cache[generic_type] = type(
                    'Error', (cls,), {})
            if error_type not in cls._class_cache:
                cls._class_cache[error_type] = type(
                    'Error', (cls._class_cache[generic_type],), {})
            return cls._class_cache[error_type]

        # `Error.TypeError` => the cached error class.
        __getattr__ = _get_error_cls

        # `Error['exceptions.TypeError']` => the cached error class.
        __getitem__ = _get_error_cls

    def __new__(cls, request, response):
        return super(Error, cls).__new__(
            cls._get_error_cls(response['error']['type']),
            request, response)

    def __init__(self, request, response):
        self.request = request
        self.response = response

    def __repr__(self):
        return 'Error[%s]' % (self.type,)

    def __str__(self):
        return self.message

    @property
    def id(self):
        return self.request['id']

    @property
    def method(self):
        return self.request['method']

    @property
    def params(self):
        return self.request['params']

    @property
    def type(self):
        return self.response['error']['type']

    @property
    def message(self):
        return self.response['error']['message']

    @property
    def args(self):
        return self.response['error']['args']


class Client(object):

    def __init__(self, addr, context=None):
        self.context = context or zmq.Context.instance()
        self.addr = addr
        self.socket = self.context.socket(zmq.REQ)
        self.socket.connect(self.addr)

    def __getattr__(self, method):
        return self[method]

    def __getitem__(self, method):
        return ClientMethod(self, method)

    def _process_response(self, request, response):
        if not response['error']:
            return response['result']
        raise Error(request, response)

    def __call__(self, method, *params):
        request = {"id": get_uuid(),
                   "method": method,
                   "params": params}
        self.socket.send(BSON.encode(request))
        return self._process_response(request,
                                      BSON(self.socket.recv()).decode())


class ClientMethod(object):

    def __init__(self, client, method):
        self.__client = client
        self.__method = method

    def __getattr__(self, attr):
        return self[attr]

    def __getitem__(self, item):
        return ClientMethod(self.__client, '%s.%s' % (self.__method, item))

    def __call__(self, *args, **kwargs):
        return self.__client(self.__method, *args, **kwargs)


def get_uuid():
    return str(uuid.uuid4()).replace('-', '')



"""
Common exceptions throughout the ZRPC server.

Note that :class:`zrpc.client.Error` doesn't live here; this module houses
server-specific exceptions.
"""

class ZRPCError(Exception):
    """A generic error somewhere in the ZRPC server."""
    pass


class MissingMethod(Exception):
    """The requested method was not defined on the server."""
    pass



from __future__ import with_statement

from contextlib import closing, nested

import logbook
import zmq

from zrpc.concurrency import DummyCallback


logger = logbook.Logger('zrpc.loadbal')


class LoadBalancer(object):

    """
    Load balance to multiple ZRPC servers using a ``zmq.QUEUE`` device.

    There are two ways to configure this load balancer: the ZeroMQ docs refer
    to these as the 'proxy' and 'broker' models, although I like to think of
    them as asymmetric and symmetric, respectively.

    In the proxy model, the load balancer binds to an input socket and connects
    to a number of output sockets downstream. This means no changes are
    required to the servers that are being load-balanced, but the list of
    servers is fixed; adding or removing servers from the cluster requires that
    you reconfigure and restart the load balancer.

    To create a proxy load balancer, pass a list of addresses as the `output`
    argument:

        >>> loadbal = LoadBalancer('inproc://zrpc',
        ...     ('inproc://worker0', 'inproc://worker1', ...))
        >>> loadbal.run()

    In the broker model, the load balancer binds to two sockets; one for input
    and one for output. This means the ZRPC servers now have to connect into
    the load balancer instead, but they can come and go dynamically without
    restarting the load balancer.

    To create a broker load balancer, pass a single address as the `output`
    argument:

        >>> loadbal = LoadBalancer('inproc://zrpc', 'inproc://zrpc-workers')
        >>> loadbal.run()
    """

    def __init__(self, input, output, context=None):
        self.context = context or zmq.Context.instance()
        self.input = input
        self.output = output

    def run(self, callback=DummyCallback(), device=zmq.device):

        """
        Run the load balancer.

        :param callback:
            A :class:`~zrpc.concurrency.Callback` which will be called with the
            input and output sockets once they have been successfully connected
            or bound.
        """

        with callback.catch_exceptions():
            logger.debug("Listening for requests on {0!r}", self.input)
            input_socket = self.context.socket(zmq.XREP)
            input_socket.bind(self.input)

            output_socket = self.context.socket(zmq.XREQ)
            if hasattr(self.output, '__iter__'):
                logger.debug("Connecting to {0} workers", len(self.output))
                for node in self.output:
                    output_socket.connect(node)
            else:
                logger.debug("Listening for workers on {0!r}", self.output)
                output_socket.bind(self.output)
        callback.send((input_socket, output_socket))

        with nested(logger.catch_exceptions(),
                    closing(input_socket),
                    closing(output_socket)):
            try:
                device(zmq.QUEUE, input_socket, output_socket)
            except zmq.ZMQError, exc:
                if exc.errno == zmq.ETERM:
                    pass
                else:
                    raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from distribute_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages
import os.path as p

VERSION = open(p.join(p.dirname(p.abspath(__file__)), 'VERSION')).read().strip()

setup(
    name='zrpc',
    version=VERSION,
    description='Simple ZeroMQ-based RPC.',
    author='Zachary Voase',
    author_email='z@dvxhouse.com',
    url='http://github.com/zacharyvoase/zrpc',
    packages=find_packages(where='lib'),
    package_dir={'': 'lib'},
    install_requires=[
        'pyzmq>=2.1.10',
        'Logbook==0.3',
        'PyMongo==2.1',
    ],
)
